/*
	Stonehenge robot arm program
	
		Based on Philippe E. Hurbain's wonderful design
		
		author: R. Hansen Aug. 2010
		
		The Lego NXT 2.0 kit comes with a total of 12 small balls,
		three each of four different colors.  If any combination of 1 to 8
		of these balls are placed on the feed rails, this program can sort them
		in "rainbow order".  This has been verified through simulation.
		It should be easy to improve the program to remove the three of each
		color limit and perhaps the number of colors limit.

	todo:
	
	 ) In state 100, don't move a 4th ball to the stands unless it's > lowest stand ball
	 ) get removeExcessBalls working correctly
	 ) try to improve 1-7 ball sorts
	 ) try to get maxballs >= 5 working
*/

#define DO_SIM

#ifdef DO_SIM
#include "nxc.h"
#endif

#define MAX_BALLS	8

int loopCnt;

long timeNow, startTime;

int sortState, oldSortState;
int iWhereAmI;
bool armIsUP;
bool armIsOpen;
int occuppied[] = { 0, 1, 0, 0, 0, 0, 0, 0, 0 };
int whichColor[] = { -1, -1, -1, -1,  -1, -1, -1, -1, -1 };
int theBallColor = -1;
int theHighestColorFound = -1;
int numBallsSkipped = 0;
int numSortedBallsOnRail = 0;
int numBallsOnStands = 0;
int numRedBallsFound = 0;

int numBallsFound  = 0;

//int stationAngles[] = { -10, -63, -90, -127,  -150, -174, -207,  -240, -300 };
//int stationAngles[] = { -10, -63, -90, -127,  -150, -173, -207,  -240, -300 };
int stationAngles[] = {   -10, -63, -89, -127,  -150, -173, -207,  -240, -300 };



#define MIN_RAIL	1
#define MAX_RAIL    8
#define NUM_STANDS  4

int theRail[] = { 0, 0, 0, 0,  0, 0, 0, 0 };
int theStands[] = { -1, -1, -1, -1 };

#ifdef DO_SIM
typedef
#endif
struct theRailsT {
	bool valid;
	int lastBallRemoved;
#ifdef DO_SIM
} theRailsT;
#else
};
#endif

theRailsT theRails[MAX_BALLS+1];
int trueNumOfBalls = -1;

string colorTable[] = { "N", "b", "B", "G", "Y", "R", "w" };

#define RAIL_DROPOFF	0
#define RAIL_PICKUP		1
#define STAND1			2
#define STAND2			3
#define C_SENSOR		4
#define STAND3			5
#define STAND4			6
#define BIN1			7
#define BIN2			8

byte rotatorMotor = OUT_A;
byte grabberMotor = OUT_B;
byte lifterMotor  = OUT_C;


void init();
void initLogfileInfo();
safecall void makeLogEntry( byte type, int value );
void parkMotor( byte motor, int power, int direction );
int grab();
void release();
void raiseTheArm();
void lowerTheArm();
void rotateArmTo( int iStation );
int grabaBall();
int pickupBallAt( int iStation );
int dropoffBallAt( int iStation );
void checkBallColor();
void showTheRail();
void dumpBallStatus();
void shiftTheRailPast( int thisColor );
void removeExcessBalls( int iColor );
void findNumSortedBallsOnRail( int iColor );
int isBallStuck();
int wasThisBallStuck();

void initTheRails();
void updateTheRails( int inOrOut );
void checkTheRails( int expectedBallColor );
bool shortCutIsSafe( int grabbedBallColor );
bool isTheRailSorted();
bool isOKToMoveToRail();

void juggle();
void sort();
int getColor();
int findFreeStand();
int findLowestStand();
int findHighestStand();
int findHighestStandColor();
int findHighestRail();



#ifdef DO_SIM
stonehenge()
{
	extern int isDone;				// New
#else
task main()
{
#endif

	init();
	
	while( 1 ) {
#ifdef DO_SIM
		if( isDone ) break;			// New
#endif
		Wait( 50 );
		loopCnt++;
		makeLogEntry( 3, loopCnt );
		if( loopCnt > 40 ) {
			makeLogEntry( 4, 30 );
			Stop( 1 );
		}

#if 0
		juggle();
#else
		sort();
		dumpBallStatus();
#endif
	}
}


void
init()
{
	int i, iTo;

	loopCnt		 =   0;
	sortState    = 100;
	oldSortState =  -1;
	iWhereAmI    =  -1;
	
	numBallsFound   =  0;
	numBallsSkipped =  0;
	trueNumOfBalls  = -1;
	numSortedBallsOnRail  = 0;
	numBallsOnStands   = 0;
	numRedBallsFound   = 0;
	theHighestColorFound = -1;

	for( i=0; i<4; i++ ) theStands[i] = -1;
	for( i=0; i<8; i++ ) theRail[i]   =  0;
	for( i=0; i<9; i++ ) {
		occuppied[i] =   0;
		whichColor[i] = -1;
	}
	occuppied[1] = 1;

	timeNow = CurrentTick();
	startTime = timeNow;
	
	initLogfileInfo();
	
	parkMotor( lifterMotor,  35, 1 );			// Reset the lifter
	parkMotor( grabberMotor, 20, 1 );			// Reset the grabber
	parkMotor( rotatorMotor, 30, 1 );			// Reset the rotator
	
	armIsUP   = TRUE;
	armIsOpen = FALSE;
	
	initTheRails();

#if 0
	iTo = 3;
	for( i=0; i<4; i++ ) {
		pickupBallAt( iTo );
		dropoffBallAt( iTo );
	}
	
	Stop( 1 );
#endif

#if 0
	iTo = 0;
	rotateArmTo( iTo );
	for( i=0; i<4; i++ ) {
		lowerTheArm();
		raiseTheArm();
	}
	
	Stop( 1 );
#endif

#if 0
	iTo = 0;
	rotateArmTo( iTo );
	release();
	lowerTheArm();
	for( i=0; i<8; i++ ) {
		grab();
		release();
	}
	
	Stop( 1 );
#endif

}


void
initTheRails()
{
	int i;
	
	for( i=MIN_RAIL; i<=MAX_RAIL; i++ ) {
		theRails[i].valid = TRUE;
		theRails[i].lastBallRemoved = -1;
	}
}


void
updateTheRails( int inOrOut )
{
	int inx, iRails;

	if( inOrOut == RAIL_PICKUP ) {
		
		for( iRails=MIN_RAIL; iRails<=MAX_RAIL; iRails++ ) {
			inx = MAX_BALLS - iRails - numBallsSkipped;
			if( theRails[iRails].valid  &&  inx >= 0 ) {
				theRails[iRails].lastBallRemoved = theRail[inx];
//				makeLogEntry( 62, 100*iRails + 10*inx + 1 + theRails[iRails].lastBallRemoved );
			}
		}

		theBallColor = theRail[0];
		for( inx=0; inx<7; inx++ ) theRail[inx] = theRail[inx+1];
		theRail[7] = -1;
		showTheRail();
	}
	
	if( inOrOut == RAIL_DROPOFF ) {
		occuppied[RAIL_PICKUP] = 1;
		for( inx=0; inx<7; inx++ ) if( theRail[inx] == -1 ) break;
		theRail[inx] = theBallColor;
		showTheRail();
	}
	

}


void
checkTheRails( int expectedBallColor )
{
	int iRails, iValid, iColor;
	
	if( expectedBallColor > 0  &&  expectedBallColor != theBallColor ) {
		shiftTheRailPast( theBallColor );
	}

	for( iRails=MIN_RAIL; iRails<=MAX_BALLS; iRails++ ) {
		if( theRails[iRails].valid ) {
			if( numBallsSkipped > MAX_BALLS-iRails ) {
				theRails[iRails].valid = 0;
				makeLogEntry( 64, iRails*10 + numBallsSkipped );
				continue;
			}
			if( theBallColor != theRails[iRails].lastBallRemoved  &&  theRails[iRails].lastBallRemoved != 0 ) {
				theRails[iRails].valid = 0;
				makeLogEntry( 60, iRails*100 + theBallColor*10 + theRails[iRails].lastBallRemoved );
			}
		}
	}

	iValid = -1;
	for( iRails=MIN_RAIL; iRails<=MAX_RAIL; iRails++ ) {
		if( theRails[iRails].valid ) {
			if( iValid > 0 ) return;
			iValid = iRails;
		}
	}

	trueNumOfBalls = iValid;
	iColor = (expectedBallColor > 0 ) ? theBallColor : -1 ;
//	removeExcessBalls( expectedBallColor );
	removeExcessBalls( iColor );
	makeLogEntry( 61, iValid );
}


void
removeExcessBalls( int iColor )
{
	int i, j, it, iSkip, lastColor, numBallsOnRail;
	
	if( trueNumOfBalls <= 0 ) return;
	if( trueNumOfBalls == MAX_BALLS ) return;

	numBallsOnRail = 0;
	for( i=0; i<MAX_BALLS; i++ ) if( theRail[i] > 0 ) numBallsOnRail++;
	iSkip = numBallsOnRail + numBallsOnStands - trueNumOfBalls;
	if( iColor > 0 ) iSkip++;
	it = 100*iSkip + 10*numBallsOnRail + numBallsOnStands;
	makeLogEntry( 66, it );
	if( iSkip > 0 ) {
		for( i=0; i<MAX_BALLS-iSkip; i++ )
			theRail[i] = theRail[i+iSkip];
		theRail[MAX_BALLS-1] = -1;
	}

	findNumSortedBallsOnRail( iColor );
	showTheRail();
	dumpBallStatus();
}


void
findNumSortedBallsOnRail( int iColor )
{
	int i, it, highestStandColor;

	highestStandColor = findHighestStandColor();

//	printf("reb: --------------------\n");
	it = 0;
	for( i=0; i<MAX_BALLS; i++ ) {	
//		printf("reb: it %d, tr[%d] %d, hsc %d, ebc %d\n", it, i, theRail[i], highestStandColor, expectedBallColor );
		if( theRail[i] < 0 ) break;
		if( theRail[i] < highestStandColor ) {
			it = 0;
			continue;
		}
		if( iColor > 0  &&  theRail[i] < iColor ) {
			it = 0;
			continue;
		}
		if( i == MAX_BALLS-1 ) {
			it++;
			break;
		}
//		if( theRail[i+1] ) break;
		if( theRail[i] >= theRail[i+1] ) {
			it++;
		} else {
			it = 0;
		}
	}
//	it = 0;
	if( it != numSortedBallsOnRail ) {
//		printf("Reset nsbor from %d to %d\n", numSortedBallsOnRail, it );
		numSortedBallsOnRail = it;
	}
}


int
isBallStuck()
{
	int inx, inx2, numBallsOnRail, inxTopRailBall, highestRailColor;
	int numFreeStands;
	
	numFreeStands = NUM_STANDS - numBallsOnStands;
	if( numFreeStands > 0 ) return( 0 );

	numBallsOnRail = trueNumOfBalls - numBallsOnStands;
	highestRailColor = -1;

	for( inx=0; inx<numBallsOnRail; inx++ ) {
		if( highestRailColor > theRail[inx] ) continue;
		highestRailColor = theRail[inx];
		inxTopRailBall = inx;
	}
	
	for( inx=0; inx<inxTopRailBall; inx++ ) {
		for( inx2=inx+1; inx<inxTopRailBall; inx2++ ) {
			if( theRail[inx] >= theRail[inx2] ) continue;
			return( 1 );
			break;
		}
	}

	return( 0 );
}


int
wasThisBallStuck()
{
	int inx, numBallsOnRail, inxTopRailBall;
	
	if( theRail[0] == INPUT_BLUECOLOR ) return( 0 );
	
	inxTopRailBall = findHighestRail();
	
	if( inxTopRailBall == 1 ) return( 1 );
	
	numBallsOnRail = trueNumOfBalls - numBallsOnStands;
	
	if( theRail[0] > theRail[numBallsOnRail] - 1 ) return( 1 );
/*	
	for( inx=inxTopRailBall; inx<MAX_RAIL; inx++ )
		if( theRail[inx] == theRail[inxTopRailBall] ) return( 1 );
*/
	for( inx=0; inx<inxTopRailBall-1; inx++ ) {
		if( theRail[0] >= theRail[inx] ) continue;
		return( 1 );
	}

	return( 0 );
}


void
showTheRail()
{
	int i, inxColor;
	string theColor;

	for( i=0; i<8; i++ ) {
		inxColor = theRail[i];
		if( inxColor < 0  ||  inxColor > 6 )
			theColor = "x";
		else
			theColor = colorTable[inxColor];

//		TextOut( 10*i, LCD_LINE4, "  " );
		TextOut( 10*i, LCD_LINE4, theColor );
	}
}


void
shiftTheRailPast( int thisColor )
{
	int i, j, lastColor;

	for( i=0; i<7; i++ ) {
		if( theRail[0] <= 0 ) break;
		lastColor = theRail[0];
		for( j=0; j<7; j++ ) {
			theRail[j] = theRail[j+1];
			theRail[7] = -1;
		}
		numBallsSkipped++;
		makeLogEntry( 65, numBallsSkipped );
		if( thisColor == lastColor ) break;
	}

	showTheRail();
	dumpBallStatus();
}


void
dumpBallStatus()
{
	int i;

	for( i=0; i<8; i++ ) makeLogEntry( 36, (i<<8) | (theRail[i]&0xFF) );
	for( i=0; i<4; i++ ) makeLogEntry( 37, (i<<8) | (theStands[i]&0xFF) );
}

										// Sort 1 to 8 balls found on the rail
										// back onto the rail in rainbow order
int highestColor   = -1;
int numRailRotate;


void
sort()
{
	int i, it, iscis, iWhereFrom, iWhereTo, iWhere, expectedBallColor;
	int theHighestStandColor;
	long now;

	if( sortState != oldSortState ) {
		oldSortState = sortState;
		makeLogEntry( 1, sortState );
	}

	timeNow = CurrentTick();
	
	makeLogEntry( 33, numBallsFound );
	
	switch( sortState ) {

											// Find the top red and yellow balls
	case 100:
		if( !pickupBallAt( RAIL_PICKUP ) ) {
			sortState = 1000;
			return;
		}
		numBallsFound++;
		expectedBallColor = theBallColor;
		checkBallColor();
		checkTheRails( expectedBallColor );
		if( theBallColor <= INPUT_GREENCOLOR ) {
			iWhereTo = RAIL_DROPOFF;
		} else {
			iWhereTo = findFreeStand();
		}
		dropoffBallAt( iWhereTo );
		
		if( numBallsFound + numRedBallsFound >= MAX_BALLS  &&  numRedBallsFound > 0 ) {
			sortState = 300;
			break;
		}
		
		if( numBallsOnStands >= 4 ) sortState = 200;
		if( numBallsFound == MAX_BALLS ) sortState = 400;
		if( trueNumOfBalls > 0 ) sortState = 500;
		if( numBallsFound >= 4  &&  numBallsOnStands == 0 ) sortState = 110;
		break;
	
										// Move all remaining balls onto the stands
	case 110:
		if( !pickupBallAt( RAIL_PICKUP ) ) {
			sortState = 1000;
			return;
		}
		numBallsFound++;
		expectedBallColor = theBallColor;
		checkBallColor();
		checkTheRails( expectedBallColor );
		iWhereTo = findFreeStand();
		dropoffBallAt( iWhereTo );
		if( numBallsOnStands < 4 ) break;
		sortState = 410;
		break;

										// Move lowest stand ball to the rail
	case 200:
		iWhereFrom = findLowestStand();
		pickupBallAt( iWhereFrom );
		dropoffBallAt( RAIL_DROPOFF );
		sortState = 100;
		break;

										// Move the red balls from the stands to the rail
	case 300:
		iWhereFrom = findHighestStand();
		pickupBallAt( iWhereFrom );
		dropoffBallAt( RAIL_DROPOFF );
		numSortedBallsOnRail++;
		if( findHighestStandColor() == INPUT_REDCOLOR ) break;
		sortState = 400;
		break;

										// Move rest of new rail balls to stands or
										// back to the rail
	case 400:
		if( isTheRailSorted() ) {
			sortState = 1000;
			break;
		}
		//printf("400: nsbor %d\n", numSortedBallsOnRail );
/*
		if( isOKToMoveToRail() ) {
			iWhereFrom = findHighestStand();
			pickupBallAt( iWhereFrom );
			dropoffBallAt( RAIL_DROPOFF );
			numSortedBallsOnRail++;
			break;
		}
*/
		if( !pickupBallAt( RAIL_PICKUP ) ) {
			trueNumOfBalls = numBallsOnStands;
			sortState = 1000;
			break;
		}
		numBallsFound++;
		expectedBallColor = theBallColor;
		checkBallColor();
		checkTheRails( expectedBallColor );					//printf("400:nsbor %d\n", numSortedBallsOnRail );
//		if( theBallColor >= findHighestStandColor() ) {
		if( theBallColor == INPUT_REDCOLOR ) {
			dropoffBallAt( RAIL_DROPOFF );
			if( expectedBallColor <= 0 ) {
				numSortedBallsOnRail++;						//printf("400: ebc %d\n", expectedBallColor );
				if( trueNumOfBalls > 0 ) sortState = 500;
				break;
			};
//			numRailRotate = numSortedBallsOnRail - 1;
			if( trueNumOfBalls > 0 )
				numRailRotate = trueNumOfBalls - numBallsOnStands - 1;
			sortState = 450;
			if( numRailRotate == 0 ) sortState = 500;
			break;
		}
		if( shortCutIsSafe( theBallColor ) ) {
			iWhereTo = RAIL_DROPOFF;
			numSortedBallsOnRail++;
		} else {
			iWhereTo = findFreeStand();
		}
		if( iWhereTo < 0 ) {
			iWhereTo = RAIL_DROPOFF;
			numSortedBallsOnRail = 0;		//printf("400: set nsbor to 0\n");
		}
		dropoffBallAt( iWhereTo );
//		if( numBallsFound >= MAX_BALLS  ||  trueNumOfBalls > 0 ) {
		if( trueNumOfBalls > 0 ) {
			sortState = 500;
//			if( theRail[0] == theHighestColorFound ) sortState = 1000;
			if( isTheRailSorted() ) sortState = 1000;
			break;
		}
		if( numBallsOnStands == 4 ) {
			sortState = 410;
			break;
		}
		if( theRail[0] != INPUT_REDCOLOR ) break;
		sortState = 1000;
		break;

										// Move a ball the top stand ball to the rail
	case 410:
		iWhereFrom = findHighestStand();
		pickupBallAt( iWhereFrom );
		dropoffBallAt( RAIL_DROPOFF );
		numSortedBallsOnRail++;
		sortState = 400;
		break;

	case 450:
		//printf("450: nsbor %d, nrr %d\n", numSortedBallsOnRail, numRailRotate );
		pickupBallAt( RAIL_PICKUP );
		dropoffBallAt( RAIL_DROPOFF );
		numRailRotate--;
		if( numRailRotate > 0 ) break;
		sortState = 500;
		break;
		
										// We now known how many balls we have, where they are,
										// and there color.  Move them all to the rail
	case 500:
		findNumSortedBallsOnRail( -1 );
		//printf("500: nsbor %d\n", numSortedBallsOnRail );
//		if( theRail[0] == theHighestColorFound ) {
		if( isTheRailSorted() ) {
			sortState = 1000;
			break;
		}
#if 1
		if( isOKToMoveToRail() ) {
			iWhereFrom = findHighestStand();
		} else if( shortCutIsSafe( -1 ) ) {
			iWhereFrom = RAIL_PICKUP;
#else
		if( shortCutIsSafe( -1 ) ) {
			iWhereFrom = RAIL_PICKUP;
		} else if( isOKToMoveToRail() ) {
			iWhereFrom = findHighestStand();
#endif
		} else {
			if( numBallsOnStands == 4 ) {
				sortState = 600;
				break;
			}
			pickupBallAt( RAIL_PICKUP );
			iWhereTo = findFreeStand();
			if( iWhereTo < 0 ) {
				iWhereTo = RAIL_DROPOFF;
				numSortedBallsOnRail = 0;		//printf("500: set nsbor to 0\n");
			}
			dropoffBallAt( iWhereTo );
			break;
		}
		pickupBallAt( iWhereFrom );
		dropoffBallAt( RAIL_DROPOFF );
		numSortedBallsOnRail++;
		break;

											// Make room on the stands for low balls and rotate the rail
	case 600:
		if( findHighestRail() == 0 ) {
			pickupBallAt( RAIL_PICKUP );
			dropoffBallAt( RAIL_DROPOFF );
			sortState = 500;
			break;
		}
		if( isBallStuck() ) {
			iWhereFrom = findLowestStand();
			pickupBallAt( iWhereFrom );
			dropoffBallAt( RAIL_DROPOFF );
			break;
		}
		
//		if( theRail[0] < theRail[findHighestRail()] - 1 ) {
		if( !wasThisBallStuck() ) {
			pickupBallAt( RAIL_PICKUP );
			dropoffBallAt( RAIL_DROPOFF );
		} else {
			pickupBallAt( RAIL_PICKUP );
			iWhereFrom = findFreeStand();
			dropoffBallAt( iWhereFrom );
		}
		break;

	case 620:
		break;
		
											// Move the remaining balls from
											// the stands to the rail
	case 1000:
		iWhereFrom = findHighestStand();
		if( iWhereFrom < 0 ) {
			sortState = 9900;
			break;
		}
		if( !pickupBallAt( iWhereFrom ) ) break;
		dropoffBallAt( RAIL_DROPOFF );
		break;
											// We should be done
	case 9900:
		Wait( 2500 );
		Stop( 1 );
		break;

	default:
		makeLogEntry( 4, 100 );
		Stop( 1 );
		break;
	}
}

							// Start with 1 to 7 balls on the rail
							// Then randomly move them back to the rail, to a stand or a bin
							// Then randomly pick a ball from the rail or a stand and repeat
							// Stop when all the balls are in the bins
int iLastPlaced = -1;

void
juggle()
{
	int i, iWhich, iWhereFrom, iWhereTo, numFree, numOccuppied;
	int where[8];
	
	numOccuppied = 0;

	for( i=RAIL_PICKUP; i<BIN1; i++ )
		if( occuppied[i] ) where[numOccuppied++] = i;
		
	if( iLastPlaced >= 0  &&  numOccuppied > 1 ) {
		for( i=0; i<numOccuppied; i++ ) {
			if( where[i] != iLastPlaced ) continue;
			makeLogEntry( 41, i );
			for( ; i<numOccuppied-1; i++ ) {
				where[i] = where[i+1];
			}
			numOccuppied--;
			break;
		}
	}
	
	makeLogEntry( 30, numOccuppied );
		
	if( numOccuppied <= 0 ) {
		TextOut( 0, LCD_LINE1, "No balls left" );
		makeLogEntry( 4, 2 );
		Wait( 5000 );
		Stop( 1 );
	}
		
	iWhich = Random( numOccuppied );
	makeLogEntry( 31, iWhich );
	
	if( iWhich < 0  ||  iWhich >= numOccuppied ) {
		TextOut( 0, LCD_LINE1, "bad iWhich" );
		makeLogEntry( 4, 2 );
		Wait( 5000 );
		Stop( 1 );
	}
	iWhereFrom = where[iWhich];
	
	if( iWhereFrom < RAIL_DROPOFF  ||  iWhereFrom > BIN2 ) {
		TextOut(  0, LCD_LINE1, "bad iWhereFrom" );
		TextOut( 75, LCD_LINE1, "    " );
		NumOut(  75, LCD_LINE1, iWhereFrom );
		makeLogEntry( 4, 3 );
		Wait( 5000 );
		Stop( 1 );
	}
	
	if( !pickupBallAt( iWhereFrom ) ) {
		occuppied[iWhereFrom] = 0;
		return;
	}
	
	if( iWhereFrom != RAIL_PICKUP ) occuppied[iWhereFrom] = 0;
	
	numFree = 0;
	for( i=RAIL_DROPOFF; i<=BIN2; i++ ) {
		if( i == RAIL_PICKUP ) continue;
		if( i == C_SENSOR ) continue;
		if( i == iWhereFrom ) continue;
		if( !occuppied[i] ) where[numFree++] = i;
	}
		
	if( numFree <= 0  ||  numFree > 8 ) {
		TextOut(  0, LCD_LINE1, "bad numFree" );
		TextOut( 75, LCD_LINE1, "    " );
		NumOut(  75, LCD_LINE1, numFree );
		makeLogEntry( 4, 4 );
		Wait( 5000 );
		Stop( 1 );
	}
	
	iWhich = Random(numFree);
	
	if( iWhich < 0  ||  iWhich >= numFree ) {
		TextOut( 0, LCD_LINE1, "bad iWhich" );
		makeLogEntry( 4, 5 );
		Wait( 5000 );
		Stop( 1 );
	}
	iWhereTo = where[iWhich];
	
	if( iWhereTo < 0  ||  iWhereTo > 8 ) {
		TextOut(  0, LCD_LINE1, "bad iWhereTo" );
		TextOut( 75, LCD_LINE1, "    " );
		NumOut(  75, LCD_LINE1, iWhereTo );
		makeLogEntry( 4, 5 );
		Wait( 5000 );
		Stop( 1 );
	}

	dropoffBallAt( iWhereTo );
	
	if( iWhereTo == RAIL_DROPOFF ) iWhereTo = RAIL_PICKUP;
	if( iWhereTo < BIN1 ) occuppied[iWhereTo] = 1;
	
	iLastPlaced = iWhereTo;
	if( iLastPlaced == RAIL_PICKUP ) iLastPlaced = -1;
	makeLogEntry( 40, iLastPlaced );
}


bool
shortCutIsSafe( int grabbedBallColor )
{
	int it, inx, itnob, iStart, theColor, iHighestStand, iHighestStandColor;

	theColor = (grabbedBallColor > 0) ? grabbedBallColor : theRail[0];
	itnob = (trueNumOfBalls > 0) ? trueNumOfBalls : 8;
	it = itnob - numBallsOnStands - numSortedBallsOnRail;
	iStart = 1;
	if( grabbedBallColor > 0 ) {
		it--;
		iStart = 0;
	}
	makeLogEntry( 73, 1000*it + 100*itnob + 10*numBallsOnStands + numSortedBallsOnRail );

	for( inx=iStart; inx<it; inx++ ) {
//		printf("scis: tr[%d] %d, tc %d\n", inx, theRail[inx], theColor );
		if( theRail[inx] <= 0 ) return( FALSE );
		if( theColor >= theRail[inx] ) continue;
		return( FALSE );
	}
#if 1
	it = itnob - numBallsOnStands;
	if( grabbedBallColor > 0 ) it--;
#if 0
	printf("scis: (%6d) it-1 %d, tc %d, tr %d, gc %d\n", CurrentTick(), it-1, theColor, theRail[it-1], grabbedBallColor );
	for( inx=0; inx<it; inx++ ) printf(" %d", theRail[inx]);
	printf("\n");
#endif
	if( theColor > theRail[it-1] ) return( FALSE );
#endif
	iHighestStandColor = findHighestStandColor();
	if( theColor < iHighestStandColor ) return( FALSE );

	makeLogEntry( 74, iHighestStandColor );
	return( TRUE );
}


bool
isTheRailSorted()
{
	int iBall, iHighestStandColor;

	iHighestStandColor = findHighestStandColor();
	makeLogEntry( 75, iHighestStandColor );
	
	if( theRail[0] < 0 ) return( FALSE );

	for( iBall=0; iBall<MAX_BALLS; iBall++ ) {
		if( theRail[iBall] < 0 ) break;
		if( theRail[iBall] < iHighestStandColor ) return( FALSE );
		if( iBall < MAX_BALLS-1  &&  theRail[iBall] < theRail[iBall+1] ) return( FALSE );
	}
	
	return( TRUE );
}


bool
isOKToMoveToRail()
{
	int it, inx, itnob, iWhere, highestStandColor;

	iWhere = findHighestStand();
	if( iWhere < STAND1  ||  iWhere > STAND4 ) return( FALSE );
	highestStandColor = whichColor[iWhere];

	itnob = trueNumOfBalls;
	if( itnob < 0 ) itnob = 8;
	it = itnob - numBallsOnStands - numSortedBallsOnRail;
#if 0
	printf("isOKToMoveToRail: it %d, tnob %d, nbos %d, ntbor %d\n",
		it, trueNumOfBalls, numBallsOnStands, numSortedBallsOnRail );
#endif
	makeLogEntry( 77, 1000*it + 100*itnob + 10*numBallsOnStands + numSortedBallsOnRail );
	for( inx=0; inx<it; inx++ ) {
		if( theRail[inx] <= highestStandColor ) continue;
//		printf("return false\n");
		return( FALSE );
	}
	return( TRUE );
}


int
findHighestRail()
{
	int inx, highestRailColor, inxHighestRailColor;
	
	highestRailColor = -1;

	for( inx=0; inx<MAX_RAIL; inx++ ) {
		if( theRail[inx] <= 0 ) break;
		if( highestRailColor >= theRail[inx] ) continue;
		highestRailColor = theRail[inx];
		inxHighestRailColor = inx;
	}
//	printf("fhr: %d\n", inxHighestRailColor );
	return( inxHighestRailColor );
}


int
findFreeStand()
{
	int i, iStand;
	
	for( i=0; i<4; i++ ) {
		iStand = STAND1 + i;
		if( i >= 2 ) iStand++;
		if( !occuppied[iStand] ) return( iStand );
	}
	return( -1 );
}


int
findLowestStand()
{
	int i, iStand, iWhich, lowestColor;
	
	lowestColor = INPUT_WHITECOLOR;
	iWhich = -1;

	for( i=0; i<4; i++ ) {
		iStand = STAND1 + i;
		if( i >= 2 ) iStand++;
		if( !occuppied[iStand] ) continue;
		if( whichColor[iStand] >= lowestColor ) continue;
		lowestColor = whichColor[iStand];
		iWhich = iStand;
	}
	return( iWhich );
}


int
findHighestStand()
{
	int i, iStand, iWhich, highestColor;
	
	highestColor = INPUT_BLACKCOLOR;
	iWhich = -1;

	for( i=0; i<4; i++ ) {
		iStand = STAND1 + i;
		if( i >= 2 ) iStand++;
//		makeLogEntry( 76, 100*i + 10*iWhich + highestColor );
		if( whichColor[iStand] <= highestColor ) continue;
		highestColor = whichColor[iStand];
		iWhich = iStand;
//		makeLogEntry( 71, 10*iWhich + highestColor );
	}
//	makeLogEntry( 71, 10*iWhich + highestColor );
	return( iWhich );
}


int
findHighestStandColor()
{
	int i, iStand, iWhere, highestColor;
	
	highestColor = INPUT_BLACKCOLOR;
	iWhere = findHighestStand();
	
	if( iWhere >= STAND1 ) highestColor = whichColor[iWhere];
	
	return( highestColor );
}



int colorSensorPort = S2;
ColorSensorReadType csr;


int
getColor()
{
	int red, green, blue, total;

	SetSensorColorFull( colorSensorPort	);
	csr.Port = colorSensorPort;
	SysColorSensorRead( csr );
	
	red   = csr.NormalizedArray[0];
	green = csr.NormalizedArray[1];
	blue  = csr.NormalizedArray[2];
	
	SetSensorColorNone( colorSensorPort );
	
	makeLogEntry( 51, red );
	makeLogEntry( 52, green );
	makeLogEntry( 53, blue );
	
	total = red + green + blue;
	
	if( total < 40 ) return( INPUT_BLACKCOLOR );
	
	if( blue/2    > red    &&  2*blue/3  > green ) return( INPUT_BLUECOLOR );
	
	if( 2*green/3 > red    &&  3*green/4 > blue )  return( INPUT_GREENCOLOR );
	
	if( 9*red/10  > green  &&  red/3    < green  &&  red/2 > blue ) return( INPUT_YELLOWCOLOR );
	
	if( red/3     > green  &&  red/4    > blue )   return( INPUT_REDCOLOR );
	
	return( INPUT_WHITECOLOR );
}


int
pickupBallAt( int iStation )
{
	int i, it;

	rotateArmTo( iStation );
	release();
	lowerTheArm();

	if( iStation >= STAND1  &&  iStation <= STAND4 ) {
		theBallColor = whichColor[iStation];
		occuppied[iStation]  = 0;
		whichColor[iStation] = 0;
		it = iStation - STAND1;
		if( it > 2 ) it--;
		theStands[it] = -1;
		numBallsOnStands--;
	}

	if( !grabaBall() ) {
		if( iStation == RAIL_PICKUP ) {
			occuppied[iStation] = 0;
			whichColor[iStation] = 0;
		}
		return( 0 );
	}
	raiseTheArm();
	
	if( iStation == RAIL_PICKUP ) {
		updateTheRails( RAIL_PICKUP );
	}

	return( 1 );
}


int
dropoffBallAt( int iStation )
{
	int i, it;

	rotateArmTo( iStation );
	lowerTheArm();
	release();
	raiseTheArm();

	if( iStation == RAIL_DROPOFF ) {
		updateTheRails( RAIL_DROPOFF );
	} else if( iStation >= STAND1  &&  iStation <= STAND4 ) {
		occuppied[iStation] = 1;
		whichColor[iStation] = theBallColor;
		it = iStation - STAND1;
		if( it > 2 ) it--;
		theStands[it] = theBallColor;
		numBallsOnStands++;
	}
}


void
checkBallColor()
{
	int i, j, color;

	rotateArmTo( C_SENSOR );
	for( i=0; i<2; i++ ) {
		lowerTheArm();
		for( j=0; j<2; j++ ) {
			color = getColor();
			makeLogEntry( 50, color );
			if( color > INPUT_BLACKCOLOR  &&  color < INPUT_WHITECOLOR ) break;
		}
		raiseTheArm();
		if( color > INPUT_BLACKCOLOR  &&  color < INPUT_WHITECOLOR ) break;
	}
	
	theBallColor = color;
	if( theHighestColorFound < color ) theHighestColorFound = color;
	if( color == INPUT_REDCOLOR ) {
		numRedBallsFound++;
//		printf("numRedBallsFound %d\n", numRedBallsFound );
	}
}

									// Run a motor softly until it stops
									// Reset the rotationCount to mark this as its zero position
void
parkMotor( byte motor, int power, int direction )
{
	int iPower, iStart, iNow;
	
	iPower = (direction >= 0 ) ? power : -power;
	OnFwd( motor, iPower );

	iStart = MotorRotationCount( motor );
	
	while( 1 ) {
		Wait( 100 );
		iNow = MotorRotationCount( motor );
		if( abs( iNow - iStart ) < 2 ) break;
		iStart = iNow;
	}

	Off( motor );
	
	Wait( 300 );
	
	ResetRotationCount( motor );
	
}


int
grab()
{
	int iStart, iNow;
	
//	OnFwd( grabberMotor, 15 );
//	OnFwd( grabberMotor, 20 );
	OnFwd( grabberMotor, 40 );

	iStart = MotorRotationCount( grabberMotor );
	Wait( 20 );
	
	while( 1 ) {
		Wait( 80 );
		iNow = MotorRotationCount( grabberMotor );
		if( abs( iNow - iStart ) < 1 ) break;
		iStart = iNow;
	}

	Off( grabberMotor );

	Wait( 50 );
	iNow = MotorRotationCount( grabberMotor );

	makeLogEntry( 20, iNow );
	armIsOpen = FALSE;
	
	if( iNow < -30 ) {
		makeLogEntry( 4, 20 );
		Stop( 1 );
	}

	return( iNow );
}


/*
#define MIN_BALL_THRESHOLD	-24			// Probably closed on a structure and not a ball
#define MAX_BALL_THRESHOLD   -8			// Came up empty
*/
#define MIN_BALL_THRESHOLD	-10			// Probably closed on a structure and not a ball
#define MAX_BALL_THRESHOLD    5			// Came up empty


int
grabaBall()
{
	int i, iAngle;

	for( i=0; i<2; i++ ) {
		iAngle = grab();
		
		if( iAngle < MIN_BALL_THRESHOLD ) {
			release();
			if( i > 0 ) continue;
			raiseTheArm();
			lowerTheArm();
			continue;
		}

		if( iAngle > MAX_BALL_THRESHOLD ) {
			release();
			continue;
		}
		break;
	}
	
	if( iAngle >= MIN_BALL_THRESHOLD  &&  iAngle <= MAX_BALL_THRESHOLD ) return( 1 );

	raiseTheArm();
	release();
	return( 0 );
}


void
release()
{
	int iStart, iNow;
	
	if( armIsOpen ) return;
#if 1
//	OnFwd( grabberMotor, -12 );
	OnFwd( grabberMotor, -15 );
	OnFwd( grabberMotor, -40 );

	iStart = MotorRotationCount( grabberMotor );
	Wait( 50 );
	
	while( 1 ) {
		Wait( 25 );
		iNow = MotorRotationCount( grabberMotor );
		if( abs( iNow - iStart ) < 1 ) break;
		if( iNow < -37 ) break;
		iStart = iNow;
	}

	Off( grabberMotor );
#else
	iNow = MotorRotationCount( grabberMotor );
	RotateMotorEx( grabberMotor, 12, -35-iNow, 0, FALSE, TRUE );
	RotateMotorEx( grabberMotor, 12, -38-iNow, 0, FALSE, TRUE );
	Wait( 100 );
	iNow = MotorRotationCount( grabberMotor );
#endif
	armIsOpen = TRUE;
	makeLogEntry( 21, iNow );
}


void
raiseTheArm()
{
	int iStart, iNow;
	
	OnFwd( lifterMotor, 50 );

	iStart = MotorRotationCount( lifterMotor );
#if 1
//	RotateMotorEx( lifterMotor, 50, -40-iStart, 0, FALSE, TRUE );
	RotateMotorEx( lifterMotor, 50, -35-iStart, 0, FALSE, TRUE );
	iNow = MotorRotationCount( lifterMotor );
#else
	while( 1 ) {
		Wait( 100 );
		iNow = MotorRotationCount( lifterMotor );
		if( abs( iNow - iStart ) < 1 ) break;
		iStart = iNow;
	}

	Off( lifterMotor );
#endif
	makeLogEntry( 26, iNow );
	
	armIsUP = TRUE;
}


void
lowerTheArm()
{
	int iStart, iNow, iRotateAngle;
	
	OnFwd( lifterMotor, -50 );
	Wait( 50 );
	OnFwd( lifterMotor, -25 );

	iStart = MotorRotationCount( lifterMotor );
	
	while( 1 ) {
		Wait( 50 );
		iNow = MotorRotationCount( lifterMotor );
		if( abs( iNow - iStart ) < 1 ) break;
		iStart = iNow;
	}

	Off( lifterMotor );
	
	armIsUP = FALSE;

	iNow = MotorRotationCount( lifterMotor );
	makeLogEntry( 25, iNow );
	
	if( iNow > -80 ) {
		makeLogEntry( 4, 200 );
		Stop( 1 );
	}
}


void
rotateArmTo( int iStation )
{
	int i, it, iAngle, iNow, iLast, iPower, iPrevPower, iDiff, iVel;
	int iTimeNow, iTimeLast, maxPower, maxDiff;
	float vel, deltaTime;
	
	makeLogEntry( 5, iStation );

	if( iStation < 0  ||  iStation > 8 ) {
		makeLogEntry( 4, 10 );
		Stop( 1 );
	}

	iAngle = stationAngles[iStation];
	
	iNow = MotorRotationCount( rotatorMotor );
	makeLogEntry( 6, iNow );
	
	if( !armIsUP ) {
		TextOut(  0, LCD_LINE1, "Can't rotate, arm is down" );
		Wait( 5000 );
		makeLogEntry( 4, 1 );
		Stop( 1 );
	}
	
	iDiff = iAngle - iNow;
	if( abs(iDiff) >= 1 ) {
		iTimeLast = CurrentTick();
		vel = 0.0;
		iVel     =  0;
		maxPower = 50;
		maxDiff  = 50;
		iPrevPower = 0;

		for( i=0; i<80; i++ ) {
			it = iDiff;
			if( it < -maxDiff ) it = -maxDiff;
			if( it >  maxDiff ) it =  maxDiff;
			iPower = 4  +  4 * abs(it) / 3;
			if( iVel == 0  &&  abs(iPrevPower) < 10 ) iPower += 7;
			if( iDiff < 0 ) iPower = -iPower;
			iPower -= iVel / 16;
			if( iPower < -maxPower ) iPower = -maxPower;
			if( iPower >  maxPower ) iPower =  maxPower;
			iPrevPower = iPower;
			OnFwdRegEx( rotatorMotor, iPower, OUT_REGMODE_IDLE, RESET_NONE );
			iLast = iNow;
			iNow = MotorRotationCount( rotatorMotor );
			iTimeNow = CurrentTick();
			
			deltaTime = 0.001 * (iTimeNow - iTimeLast);
			if( deltaTime <= 0.0 ) deltaTime = 1.0;
			vel = (iNow - iLast) / deltaTime;
			iVel = vel;
			iDiff = iAngle - iNow;
#if 0
			if( abs(iDiff) < maxDiff ) {
				makeLogEntry( 90, iPower );
				makeLogEntry( 91, iNow );
				makeLogEntry( 92, iVel );
				makeLogEntry( 93, iDiff );
			}
#endif
			if( abs(iDiff) < 1  &&  abs(iVel) < 50 ) break;
			if( abs(iDiff) == 1 &&  iVel == 0 ) break;
			iTimeLast = iTimeNow;
			iLast     = iNow;
			Wait( 20 );
		}
		
		OnFwd( rotatorMotor, 0 );
	}
	iNow = MotorRotationCount( rotatorMotor );
	makeLogEntry( 6, iNow );
	makeLogEntry( 7, iNow - iAngle );
	iWhereAmI = iStation;
}

#if 0
void
rotateArmTo( int iStation )
{
	int i, iAngle, iNow, iPower, iDiff;
	
	makeLogEntry( 5, iStation );

	if( iStation < 0  ||  iStation > 8 ) {
		makeLogEntry( 4, 10 );
		Stop( 1 );
	}
	iAngle = stationAngles[iStation];
	
	if( iWhereAmI == iStation ) {
		iNow = MotorRotationCount( rotatorMotor );
		makeLogEntry( 7, iNow - iAngle );
		return;
	}
	
	if( !armIsUP ) {
		TextOut(  0, LCD_LINE1, "Can't rotate, arm is down" );
		Wait( 5000 );
		makeLogEntry( 4, 1 );
		Stop( 1 );
	}
	
	iPower = 35;

	iNow = MotorRotationCount( rotatorMotor );
	
	for( i=0; i<3; i++ ) {
		iDiff = iAngle - iNow;
		if( abs(iDiff) < 2 ) break;
		
		if( i > 0 ) iPower = 2 * iPower / 3;
	
		RotateMotorEx( rotatorMotor, iPower, iDiff, 0, FALSE, TRUE );
//		RotateMotorEx( rotatorMotor, iPower,     0, 0, FALSE, TRUE );
	
//		Wait( 100 );

		OnFwd( rotatorMotor, 0 );
		iNow = MotorRotationCount( rotatorMotor );
		makeLogEntry( 6, iNow );
	}
	makeLogEntry( 7, iNow - iAngle );
	iWhereAmI = iStation;
}
#endif



byte LFH;
int bytesRemaining;

int logfileSize = 10000;


void
initLogfileInfo()
{
	DeleteFile( "logfile.dat" );
	CreateFile( "logfile.dat", logfileSize, LFH );
	
	bytesRemaining = logfileSize;
}



safecall void
makeLogEntry( byte type, int value )
{
	int time;
	long now;
	
	now = CurrentTick();
	time = now;

	Write( LFH, type );
	Write( LFH, time );
	Write( LFH, value );
	bytesRemaining -= 5;
}
